@@ -5,9 +5,10 @@
 -- file format documentation</a> for details.
 --
 -- End-users shouldn't edit this file. They can override any defaults
--- set in this file using their system-wide or user-specific configuration
--- files. Run `luarocks` with no arguments to see the locations of
--- these files in your platform.
+-- set in this file using their system-wide $LUAROCKS_SYSCONFIG file
+-- (see luarocks.site_config) or their user-specific configuration file
+-- (~/.luarocks/config.lua on Unix or %APPDATA%/luarocks/config.lua on
+-- Windows).
 
 local rawset, next, table, pairs, require, io, os, setmetatable, pcall, ipairs, package, tonumber, type, assert, _VERSION =
       rawset, next, table, pairs, require, io, os, setmetatable, pcall, ipairs, package, tonumber, type, assert, _VERSION
@@ -31,12 +32,11 @@ if not ok then
    site_config = {}
 end
 
-cfg.program_version = "2.4.1"
-cfg.program_series = "2.4"
+cfg.site_config = site_config
+
+cfg.program_version = "2.2.2"
+cfg.program_series = "2.2"
 cfg.major_version = (cfg.program_version:match("([^.]%.[^.])")) or cfg.program_series
-cfg.variables = {}
-cfg.rocks_trees = {}
-cfg.platforms = {}
 
 local persist = require("luarocks.persist")
 
@@ -70,94 +70,67 @@ end
 
 -- System detection:
 
+local detected = {}
+local system,proc
+
 -- A proper installation of LuaRocks will hardcode the system
 -- and proc values with site_config.LUAROCKS_UNAME_S and site_config.LUAROCKS_UNAME_M,
 -- so that this detection does not run every time. When it is
 -- performed, we use the Unix way to identify the system,
 -- even on Windows (assuming UnxUtils or Cygwin).
-local system = site_config.LUAROCKS_UNAME_S or io.popen("uname -s"):read("*l")
-local proc = site_config.LUAROCKS_UNAME_M or io.popen("uname -m"):read("*l")
+system = site_config.LUAROCKS_UNAME_S or io.popen("uname -s"):read("*l")
+proc = site_config.LUAROCKS_UNAME_M or io.popen("uname -m"):read("*l")
 if proc:match("i[%d]86") then
-   cfg.target_cpu = "x86"
+   proc = "x86"
 elseif proc:match("amd64") or proc:match("x86_64") then
-   cfg.target_cpu = "x86_64"
+   proc = "x86_64"
 elseif proc:match("Power Macintosh") then
-   cfg.target_cpu = "powerpc"
- else
-   cfg.target_cpu = proc
+   proc = "powerpc"
 end
 
 if system == "FreeBSD" then
-   cfg.platforms.unix = true
-   cfg.platforms.freebsd = true
-   cfg.platforms.bsd = true
+   detected.unix = true
+   detected.freebsd = true
+   detected.bsd = true
 elseif system == "OpenBSD" then
-   cfg.platforms.unix = true
-   cfg.platforms.openbsd = true
-   cfg.platforms.bsd = true
+   detected.unix = true
+   detected.openbsd = true
+   detected.bsd = true
 elseif system == "NetBSD" then
-   cfg.platforms.unix = true
-   cfg.platforms.netbsd = true
-   cfg.platforms.bsd = true
+   detected.unix = true
+   detected.netbsd = true
+   detected.bsd = true
 elseif system == "Darwin" then
-   cfg.platforms.unix = true
-   cfg.platforms.macosx = true
-   cfg.platforms.bsd = true
+   detected.unix = true
+   detected.macosx = true
+   detected.bsd = true
 elseif system == "Linux" then
-   cfg.platforms.unix = true
-   cfg.platforms.linux = true
+   detected.unix = true
+   detected.linux = true
 elseif system == "SunOS" then
-   cfg.platforms.unix = true
-   cfg.platforms.solaris = true
+   detected.unix = true
+   detected.solaris = true
 elseif system and system:match("^CYGWIN") then
-   cfg.platforms.unix = true
-   cfg.platforms.cygwin = true
-elseif system and system:match("^MSYS") then
-   cfg.platforms.unix = true
-   cfg.platforms.msys = true
-   cfg.platforms.cygwin = true
+   detected.unix = true
+   detected.cygwin = true
 elseif system and system:match("^Windows") then
-   cfg.platforms.windows = true
-   cfg.platforms.win32 = true
+   detected.windows = true
 elseif system and system:match("^MINGW") then
-   cfg.platforms.windows = true
-   cfg.platforms.mingw32 = true
-   cfg.platforms.win32 = true
-elseif system == "Haiku" then
-   cfg.platforms.unix = true
-   cfg.platforms.haiku = true
+   detected.windows = true
+   detected.mingw32 = true
 else
-   cfg.platforms.unix = true
+   detected.unix = true
    -- Fall back to Unix in unknown systems.
 end
 
--- Set order for platform overrides
-local platform_order = {
-   -- Unixes
-   unix = 1,
-   bsd = 2, 
-   solaris = 3,
-   netbsd = 4,
-   openbsd = 5,
-   freebsd = 6,
-   linux = 7,
-   macosx = 8,
-   cygwin = 9,
-   msys = 10,
-   haiku = 11,
-   -- Windows
-   win32 = 12,
-   mingw32 = 13,
-   windows = 14 }
-
 -- Path configuration:
+
 local sys_config_file, home_config_file
-local sys_config_file_default, home_config_file_default
 local sys_config_dir, home_config_dir
 local sys_config_ok, home_config_ok = false, false
 local extra_luarocks_module_dir
-sys_config_dir = site_config.LUAROCKS_SYSCONFDIR or site_config.LUAROCKS_PREFIX
-if cfg.platforms.windows then
+sys_config_dir = site_config.LUAROCKS_SYSCONFDIR
+if detected.windows then
    cfg.home = os.getenv("APPDATA") or "c:"
    sys_config_dir = sys_config_dir or "c:/luarocks"
    home_config_dir = cfg.home.."/luarocks"
@@ -169,102 +142,61 @@ else
    cfg.home_tree = (os.getenv("USER") ~= "root") and cfg.home.."/.luarocks/"
 end
 
--- Create global environment for the config files;
-local env_for_config_file = function() 
-   local e 
-   e = {
-      home = cfg.home,
-      lua_version = cfg.lua_version,
-      platforms = util.make_shallow_copy(cfg.platforms),
-      processor = cfg.target_cpu,   -- remains for compat reasons
-      target_cpu = cfg.target_cpu,  -- replaces `processor`
-      os_getenv = os.getenv, 
-      dump_env = function()
-         -- debug function, calling it from a config file will show all 
-         -- available globals to that config file
-         print(util.show_table(e, "global environment"))
-      end,
-   }
-   return e
+cfg.variables = {}
+cfg.rocks_trees = {}
+
+sys_config_file = site_config.LUAROCKS_SYSCONFIG or sys_config_dir.."/config-"..cfg.lua_version..".lua"
+local err, errcode
+sys_config_ok, err, errcode = persist.load_into_table(sys_config_file, cfg)
+
+if (not sys_config_ok) and errcode == "open" then -- file not found, so try alternate file
+   sys_config_file = sys_config_dir.."/config.lua"
+   sys_config_ok, err, errcode = persist.load_into_table(sys_config_file, cfg)
+end
+if (not sys_config_ok) and errcode ~= "open" then -- either "load" or "run"; bad config file, bail out with error
+   io.stderr:write(err.."\n")
+   os.exit(cfg.errorcodes.CONFIGFILE)
 end
 
--- Merge values from config files read into the `cfg` table
-local merge_overrides = function(overrides)
-   -- remove some stuff we do not want to integrate
-   overrides.os_getenv = nil
-   overrides.dump_env = nil
-   -- remove tables to be copied verbatim instead of deeply merged
-   if overrides.rocks_trees   then cfg.rocks_trees   = nil end
-   if overrides.rocks_servers then cfg.rocks_servers = nil end
-   -- perform actual merge
-   util.deep_merge(cfg, overrides)
-end
+local env_for_config_file = {
+   home = cfg.home,
+   lua_version = cfg.lua_version,
+   platform = util.make_shallow_copy(detected),
+   processor = proc,
+}
 
--- load config file from a list until first succesful one. Info is 
--- added to `cfg` module table, returns filepath of succesfully loaded
--- file or nil if it failed
-local load_config_file = function(list)
-   for _, filepath in ipairs(list) do
-      local result, err, errcode = persist.load_into_table(filepath, env_for_config_file())
-      if (not result) and errcode ~= "open" then
-         -- errcode is either "load" or "run"; bad config file, so error out
+if not site_config.LUAROCKS_FORCE_CONFIG then
+
+   local home_overrides, err, errcode
+   home_config_file = os.getenv("LUAROCKS_CONFIG_" .. version_suffix) or os.getenv("LUAROCKS_CONFIG")
+   if home_config_file then
+      home_overrides, err, errcode = persist.load_into_table(home_config_file, env_for_config_file)
+   else
+      home_config_file = home_config_dir.."/config-"..cfg.lua_version..".lua"
+      home_overrides, err, errcode = persist.load_into_table(home_config_file, env_for_config_file)
+      if (not home_overrides) and (not errcode == "run") then
+         home_config_file = home_config_dir.."/config.lua"
+         home_overrides, err, errcode = persist.load_into_table(home_config_file, env_for_config_file)
+      end
+   end
+   if home_overrides then
+      home_config_ok = true
+      if home_overrides.rocks_trees then
+         cfg.rocks_trees = nil
+      end
+      if home_overrides.rocks_servers then
+         cfg.rocks_servers = nil
+      end
+      util.deep_merge(cfg, home_overrides)
+   else
+      home_config_ok = home_overrides
+      if errcode ~= "open" then
          io.stderr:write(err.."\n")
          os.exit(cfg.errorcodes.CONFIGFILE)
       end
-      if result then
-         -- succes in loading and running, merge contents and exit
-         merge_overrides(result)
-         return filepath
-      end
-   end
-   return nil -- nothing was loaded
-end
-
-
--- Load system configuration file
-do 
-   sys_config_file_default = sys_config_dir.."/config-"..cfg.lua_version..".lua"
-   sys_config_file = load_config_file({
-      site_config.LUAROCKS_SYSCONFIG or sys_config_file_default,
-      sys_config_dir.."/config.lua",
-   })
-   sys_config_ok = (sys_config_file ~= nil)
-end
-
--- Load user configuration file (if allowed)
-if not site_config.LUAROCKS_FORCE_CONFIG then
-  
-   home_config_file_default = home_config_dir.."/config-"..cfg.lua_version..".lua"
-   
-   local config_env_var   = "LUAROCKS_CONFIG_" .. version_suffix
-   local config_env_value = os.getenv(config_env_var)
-   if not config_env_value then
-      config_env_var   = "LUAROCKS_CONFIG"
-      config_env_value = os.getenv(config_env_var)
-   end
-   
-   -- first try environment provided file, so we can explicitly warn when it is missing
-   if config_env_value then 
-      local list = { config_env_value }
-      home_config_file = load_config_file(list)
-      home_config_ok = (home_config_file ~= nil)
-      if not home_config_ok then
-         io.stderr:write("Warning: could not load configuration file `"..config_env_value.."` given in environment variable "..config_env_var.."\n")
-      end
-   end
-
-   -- try the alternative defaults if there was no environment specified file or it didn't work
-   if not home_config_ok then
-      local list = {
-         home_config_file_default,
-         home_config_dir.."/config.lua",
-      }
-      home_config_file = load_config_file(list)
-      home_config_ok = (home_config_file ~= nil)
    end
 end
 
-
 if not next(cfg.rocks_trees) then
    if cfg.home_tree then
       table.insert(cfg.rocks_trees, { name = "user", root = cfg.home_tree } )
@@ -274,39 +206,17 @@ if not next(cfg.rocks_trees) then
    end
 end
 
-
--- update platforms list; keyed -> array
-do
-   local lst = {} -- use temp array to not confuse `pairs` in loop
-   for plat in pairs(cfg.platforms) do 
-      if cfg.platforms[plat] then  -- entries set to 'false' skipped
-         if not platform_order[plat] then
-            local pl = ""
-            for k,_ in pairs(platform_order) do pl = pl .. ", " .. k end
-            io.stderr:write("Bad platform given; "..tostring(plat)..". Valid entries are: "..pl:sub(3,-1) ..".\n")
-            os.exit(cfg.errorcodes.CONFIGFILE)
-         end
-         table.insert(lst, plat)
-      else
-         cfg.platforms[plat] = nil
-      end
-   end
-   -- platform overrides depent on the order, so set priorities
-   table.sort(lst, function(key1, key2) return platform_order[key1] < platform_order[key2] end)
-   util.deep_merge(cfg.platforms, lst)
-end
-
 -- Configure defaults:
+
 local defaults = {
 
    local_by_default = false,
+   use_extensions = false,
    accept_unknown_fields = false,
    fs_use_modules = true,
    hooks_enabled = true,
    deps_mode = "one",
    check_certificates = false,
-   perm_read = "0644",
-   perm_exec = "0755",
 
    lua_modules_path = "/share/lua/"..cfg.lua_version,
    lib_modules_path = "/lib/lua/"..cfg.lua_version,
@@ -363,7 +273,6 @@ local defaults = {
       FIND = "find",
       TEST = "test",
       CHMOD = "chmod",
-      MKTEMP = "mktemp",
 
       ZIP = "zip",
       UNZIP = "unzip -n",
@@ -375,7 +284,6 @@ local defaults = {
       OPENSSL = "openssl",
       MD5 = "md5",
       STAT = "stat",
-      TOUCH = "touch",
 
       CMAKE = "cmake",
       SEVENZ = "7z",
@@ -400,13 +308,14 @@ local defaults = {
    rocks_provided = {}
 }
 
-if cfg.platforms.windows then
-   local full_prefix = (site_config.LUAROCKS_PREFIX or (os.getenv("PROGRAMFILES")..[[\LuaRocks]]))
-   extra_luarocks_module_dir = full_prefix.."/lua/?.lua"
+if detected.windows then
+   local full_prefix = site_config.LUAROCKS_PREFIX.."\\"..cfg.major_version
+   extra_luarocks_module_dir = full_prefix.."\\lua\\?.lua"
 
    home_config_file = home_config_file and home_config_file:gsub("\\","/")
    defaults.fs_use_modules = false
-   defaults.arch = "win32-"..cfg.target_cpu 
+   defaults.arch = "win32-"..proc
+   defaults.platforms = {"win32", "windows" }
    defaults.lib_extension = "dll"
    defaults.external_lib_extension = "dll"
    defaults.obj_extension = "obj"
@@ -459,7 +368,8 @@ if cfg.platforms.windows then
    defaults.web_browser = "start"
 end
 
-if cfg.platforms.mingw32 then
+if detected.mingw32 then
+   defaults.platforms = { "win32", "mingw32", "windows" }
    defaults.obj_extension = "o"
    defaults.cmake_generator = "MinGW Makefiles"
    defaults.variables.MAKE = "mingw32-make"
@@ -468,7 +378,6 @@ if cfg.platforms.mingw32 then
    defaults.variables.LD = "mingw32-gcc"
    defaults.variables.CFLAGS = "-O2"
    defaults.variables.LIBFLAG = "-shared"
-   defaults.makefile = "Makefile"
    defaults.external_deps_patterns = {
       bin = { "?.exe", "?.bat" },
       -- mingw lookup list from http://stackoverflow.com/a/15853231/1793220
@@ -484,7 +393,7 @@ if cfg.platforms.mingw32 then
 
 end
 
-if cfg.platforms.unix then
+if detected.unix then
    defaults.lib_extension = "so"
    defaults.external_lib_extension = "so"
    defaults.obj_extension = "o"
@@ -494,6 +403,7 @@ if cfg.platforms.unix then
    defaults.variables.LUA_LIBDIR = site_config.LUA_LIBDIR or "/usr/local/lib"
    defaults.variables.CFLAGS = "-O2"
    defaults.cmake_generator = "Unix Makefiles"
+   defaults.platforms = { "unix" }
    defaults.variables.CC = "gcc"
    defaults.variables.LD = "gcc"
    defaults.gcc_rpath = true
@@ -520,41 +430,26 @@ if cfg.platforms.unix then
    defaults.web_browser = "xdg-open"
 end
 
-if cfg.platforms.cygwin then
+if detected.cygwin then
    defaults.lib_extension = "so" -- can be overridden in the config file for mingw builds
-   defaults.arch = "cygwin-"..cfg.target_cpu
+   defaults.arch = "cygwin-"..proc
+   defaults.platforms = {"unix", "cygwin"}
    defaults.cmake_generator = "Unix Makefiles"
    defaults.variables.CC = "echo -llua | xargs gcc"
    defaults.variables.LD = "echo -llua | xargs gcc"
    defaults.variables.LIBFLAG = "-shared"
 end
 
-if cfg.platforms.msys then
-   -- msys is basically cygwin made out of mingw, meaning the subsytem is unixish
-   -- enough, yet we can freely mix with native win32
-   defaults.external_deps_patterns = {
-      bin = { "?.exe", "?.bat", "?" },
-      lib = { "lib?.so", "lib?.so.*", "lib?.dll.a", "?.dll.a",
-              "lib?.a", "lib?.dll", "?.dll", "?.lib" },
-      include = { "?.h" }
-   }
-   defaults.runtime_external_deps_patterns = {
-      bin = { "?.exe", "?.bat" },
-      lib = { "lib?.so", "?.dll", "lib?.dll" },
-      include = { "?.h" }
-   }
-end
-
-
-if cfg.platforms.bsd then
+if detected.bsd then
    defaults.variables.MAKE = "gmake"
    defaults.variables.STATFLAG = "-f '%OLp'"
 end
 
-if cfg.platforms.macosx then
+if detected.macosx then
    defaults.variables.MAKE = "make"
    defaults.external_lib_extension = "dylib"
-   defaults.arch = "macosx-"..cfg.target_cpu
+   defaults.arch = "macosx-"..proc
+   defaults.platforms = {"unix", "bsd", "macosx"}
    defaults.variables.LIBFLAG = "-bundle -undefined dynamic_lookup -all_load"
    defaults.variables.STAT = "/usr/bin/stat"
    defaults.variables.STATFLAG = "-f '%A'"
@@ -567,33 +462,37 @@ if cfg.platforms.macosx then
    else
       defaults.gcc_rpath = false
    end
-   defaults.variables.CC = "env MACOSX_DEPLOYMENT_TARGET=10."..version.." gcc"
-   defaults.variables.LD = "env MACOSX_DEPLOYMENT_TARGET=10."..version.." gcc"
+   defaults.variables.CC = "export MACOSX_DEPLOYMENT_TARGET=10."..version.."; gcc"
+   defaults.variables.LD = "export MACOSX_DEPLOYMENT_TARGET=10."..version.."; gcc"
    defaults.web_browser = "open"
 end
 
-if cfg.platforms.linux then
-   defaults.arch = "linux-"..cfg.target_cpu
+if detected.linux then
+   defaults.arch = "linux-"..proc
+   defaults.platforms = {"unix", "linux"}
 end
 
-if cfg.platforms.freebsd then
-   defaults.arch = "freebsd-"..cfg.target_cpu
+if detected.freebsd then
+   defaults.arch = "freebsd-"..proc
+   defaults.platforms = {"unix", "bsd", "freebsd"}
    defaults.gcc_rpath = false
    defaults.variables.CC = "cc"
    defaults.variables.LD = "cc"
 end
 
-if cfg.platforms.openbsd then
-   defaults.arch = "openbsd-"..cfg.target_cpu
+if detected.openbsd then
+   defaults.arch = "openbsd-"..proc
+   defaults.platforms = {"unix", "bsd", "openbsd"}
 end
 
-if cfg.platforms.netbsd then
-   defaults.arch = "netbsd-"..cfg.target_cpu
+if detected.netbsd then
+   defaults.arch = "netbsd-"..proc
+   defaults.platforms = {"unix", "bsd", "netbsd"}
 end
 
-if cfg.platforms.solaris then
-   defaults.arch = "solaris-"..cfg.target_cpu
-   --defaults.platforms = {"unix", "solaris"}
+if detected.solaris then
+   defaults.arch = "solaris-"..proc
+   defaults.platforms = {"unix", "solaris"}
    defaults.variables.MAKE = "gmake"
 end
 
@@ -606,11 +505,13 @@ defaults.variables.LUA = site_config.LUA_DIR_SET and (defaults.variables.LUA_BIN
 -- Add built-in modules to rocks_provided
 defaults.rocks_provided["lua"] = cfg.lua_version.."-1"
 
-if bit32 then -- Lua 5.2+
+if cfg.lua_version >= "5.2" then
+   -- Lua 5.2+
    defaults.rocks_provided["bit32"] = cfg.lua_version.."-1"
 end
 
-if utf8 then -- Lua 5.3+
+if cfg.lua_version >= "5.3" then
+   -- Lua 5.3+
    defaults.rocks_provided["utf8"] = cfg.lua_version.."-1"
 end
 
@@ -667,23 +568,17 @@ function cfg.make_paths_from_tree(tree)
    return lua_path, lib_path, bin_path
 end
 
-function cfg.package_paths(current)
+function cfg.package_paths()
    local new_path, new_cpath, new_bin = {}, {}, {}
-   local function add_tree_to_paths(tree)
+   for _,tree in ipairs(cfg.rocks_trees) do
       local lua_path, lib_path, bin_path = cfg.make_paths_from_tree(tree)
       table.insert(new_path, lua_path.."/?.lua")
       table.insert(new_path, lua_path.."/?/init.lua")
       table.insert(new_cpath, lib_path.."/?."..cfg.lib_extension)
       table.insert(new_bin, bin_path)
    end
-   if current then
-      add_tree_to_paths(current)
-   end
-   for _,tree in ipairs(cfg.rocks_trees) do
-      add_tree_to_paths(tree)
-   end
    if extra_luarocks_module_dir then
-      table.insert(new_path, extra_luarocks_module_dir)
+     table.insert(new_path, extra_luarocks_module_dir)
    end
    return table.concat(new_path, ";"), table.concat(new_cpath, ";"), table.concat(new_bin, cfg.export_path_separator)
 end
@@ -695,26 +590,11 @@ function cfg.init_package_paths()
 end
 
 function cfg.which_config()
-   local ret = {
-      system = {
-         file = sys_config_file or sys_config_file_default,
-         ok = sys_config_ok,
-      },
-      user = {
-         file = home_config_file or home_config_file_default,
-         ok = home_config_ok,
-      }
-   }
-   ret.nearest = (ret.user.ok and ret.user.file) or ret.system.file
-   return ret
+   return sys_config_file, sys_config_ok, home_config_file, home_config_ok
 end
 
 cfg.user_agent = "LuaRocks/"..cfg.program_version.." "..cfg.arch
 
-cfg.http_proxy = os.getenv("http_proxy")
-cfg.https_proxy = os.getenv("https_proxy")
-cfg.no_proxy = os.getenv("no_proxy")
-
 --- Check if platform was detected
 -- @param query string: The platform name to check.
 -- @return boolean: true if LuaRocks is currently running on queried platform.
