@@ -1,120 +1,20 @@
 
 --- Functions related to fetching and loading local and remote files.
+--module("luarocks.fetch", package.seeall)
 local fetch = {}
+package.loaded["luarocks.fetch"] = fetch
 
 local fs = require("luarocks.fs")
 local dir = require("luarocks.dir")
-local rockspecs = require("luarocks.rockspecs")
-local signing = require("luarocks.signing")
+local type_check = require("luarocks.type_check")
+local path = require("luarocks.path")
+local deps = require("luarocks.deps")
 local persist = require("luarocks.persist")
 local util = require("luarocks.util")
-local cfg = require("luarocks.core.cfg")
+local cfg = require("luarocks.cfg")
 
-
---- Fetch a local or remote file, using a local cache directory.
--- Make a remote or local URL/pathname local, fetching the file if necessary.
--- Other "fetch" and "load" functions use this function to obtain files.
--- If a local pathname is given, it is returned as a result.
--- @param url string: a local pathname or a remote URL.
--- @param mirroring string: mirroring mode.
--- If set to "no_mirror", then rocks_servers mirror configuration is not used.
--- @return (string, nil, nil, boolean) or (nil, string, [string]):
--- in case of success:
--- * the absolute local pathname for the fetched file
--- * nil
--- * nil
--- * `true` if the file was fetched from cache
--- in case of failure:
--- * nil
--- * an error message
--- * an optional error code.
-function fetch.fetch_caching(url, mirroring)
-   local repo_url, filename = url:match("^(.*)/([^/]+)$")
-   local name = repo_url:gsub("[/:]","_")
-   local cache_dir = dir.path(cfg.local_cache, name)
-   local ok = fs.make_dir(cache_dir)
-
-   local cachefile = dir.path(cache_dir, filename)
-   local checkfile = cachefile .. ".check"
-
-   if (fs.file_age(checkfile) < 10 or
-      cfg.aggressive_cache and (not name:match("^manifest"))) and fs.exists(cachefile)
-   then
-      return cachefile, nil, nil, true
-   end
-
-   local lock, errlock
-   if ok then
-      lock, errlock = fs.lock_access(cache_dir)
-   end
-
-   if not (ok and lock) then
-      cfg.local_cache = fs.make_temp_dir("local_cache")
-      if not cfg.local_cache then
-         return nil, "Failed creating temporary local_cache directory"
-      end
-      cache_dir = dir.path(cfg.local_cache, name)
-      ok = fs.make_dir(cache_dir)
-      if not ok then
-         return nil, "Failed creating temporary cache directory "..cache_dir
-      end
-      lock = fs.lock_access(cache_dir)
-   end
-
-   local file, err, errcode, from_cache = fetch.fetch_url(url, cachefile, true, mirroring)
-
-   local fd = io.open(checkfile, "wb")
-   fd:write("!")
-   fd:close()
-
-   fs.unlock_access(lock)
-   if not file then
-      return nil, err or "Failed downloading "..url, errcode
-   end
-   return file, nil, nil, from_cache
-end
-
-local function ensure_trailing_slash(url)
-   return (url:gsub("/*$", "/"))
-end
-
-local function is_url_relative_to_rocks_servers(url, servers)
-   for _, item in ipairs(servers) do
-      if type(item) == "table" then
-         for i, s in ipairs(item) do
-            local base = ensure_trailing_slash(s)
-            if string.find(url, base, 1, true) == 1 then
-               return i, url:sub(#base + 1), item
-            end
-         end
-      end
-   end
-end
-
-local function download_with_mirrors(url, filename, cache, servers)
-   local idx, rest, mirrors = is_url_relative_to_rocks_servers(url, servers)
-
-   if not idx then
-      -- URL is not from a rock server
-      return fs.download(url, filename, cache)
-   end
-
-   -- URL is from a rock server: try to download it falling back to mirrors.
-   local err = "\n"
-   for i = idx, #mirrors do
-      local try_url = ensure_trailing_slash(mirrors[i]) .. rest
-      if i > idx then
-         util.warning("Failed downloading. Attempting mirror at " .. try_url)
-      end
-      local ok, name, from_cache = fs.download(try_url, filename, cache)
-      if ok then
-         return ok, name, from_cache
-      else
-         err = err .. name .. "\n"
-      end
-   end
-
-   return nil, err, "network"
+function fetch.is_basic_protocol(protocol, remote)
+   return protocol == "http" or protocol == "https" or protocol == "ftp" or (not remote and protocol == "file")
 end
 
 --- Fetch a local or remote file.
@@ -126,57 +26,22 @@ end
 -- resulting local filename of the remote file as the basename of the URL;
 -- if that is not correct (due to a redirection, for example), the local
 -- filename can be given explicitly as this second argument.
--- @param cache boolean: compare remote timestamps via HTTP HEAD prior to
--- re-downloading the file.
--- @param mirroring string: mirroring mode.
--- If set to "no_mirror", then rocks_servers mirror configuration is not used.
--- @return (string, nil, nil, boolean) or (nil, string, [string]):
--- in case of success:
--- * the absolute local pathname for the fetched file
--- * nil
--- * nil
--- * `true` if the file was fetched from cache
--- in case of failure:
--- * nil
--- * an error message
--- * an optional error code.
-function fetch.fetch_url(url, filename, cache, mirroring)
+-- @return string or (nil, string, [string]): the absolute local pathname for the
+-- fetched file, or nil and a message in case of errors, followed by
+-- an optional error code.
+function fetch.fetch_url(url, filename, cache)
    assert(type(url) == "string")
    assert(type(filename) == "string" or not filename)
 
    local protocol, pathname = dir.split_url(url)
    if protocol == "file" then
-      local fullname = fs.absolute_name(pathname)
-      if not fs.exists(fullname) then
-         local hint = (not pathname:match("^/"))
-                      and (" - note that given path in rockspec is not absolute: " .. url)
-                      or  ""
-         return nil, "Local file not found: " .. fullname .. hint
-      end
-      filename = filename or dir.base_name(pathname)
-      local dstname = fs.absolute_name(dir.path(".", filename))
-      local ok, err
-      if fullname == dstname then
-         ok = true
-      else
-         ok, err = fs.copy(fullname, dstname)
-      end
-      if ok then
-         return dstname
-      else
-         return nil, "Failed copying local file " .. fullname .. " to " .. dstname .. ": " .. err
-      end
-   elseif dir.is_basic_protocol(protocol) then
-      local ok, name, from_cache
-      if mirroring ~= "no_mirror" then
-         ok, name, from_cache = download_with_mirrors(url, filename, cache, cfg.rocks_servers)
-      else
-         ok, name, from_cache = fs.download(url, filename, cache)
-      end
+      return fs.absolute_name(pathname)
+   elseif fetch.is_basic_protocol(protocol, true) then
+      local ok, filename = fs.download(url, filename, cache)
       if not ok then
-         return nil, "Failed downloading "..url..(name and " - "..name or ""), from_cache
+         return nil, "Failed downloading "..url..(filename and " - "..filename or ""), "network"
       end
-      return name, nil, nil, from_cache
+      return filename
    else
       return nil, "Unsupported protocol "..protocol
    end
@@ -193,7 +58,7 @@ end
 -- @return (string, string) or (nil, string, [string]): absolute local pathname of
 -- the fetched file and temporary directory name; or nil and an error message
 -- followed by an optional error code
-function fetch.fetch_url_at_temp_dir(url, tmpname, filename, cache)
+function fetch.fetch_url_at_temp_dir(url, tmpname, filename)
    assert(type(url) == "string")
    assert(type(tmpname) == "string")
    assert(type(filename) == "string" or not filename)
@@ -214,28 +79,11 @@ function fetch.fetch_url_at_temp_dir(url, tmpname, filename, cache)
       util.schedule_function(fs.delete, temp_dir)
       local ok, err = fs.change_dir(temp_dir)
       if not ok then return nil, err end
-
-      local file, err, errcode
-
-      if cache then
-         local cachefile
-         cachefile, err, errcode = fetch.fetch_caching(url)
-
-         if cachefile then
-            file = dir.path(temp_dir, filename)
-            fs.copy(cachefile, file)
-         end
-      end
-
-      if not file then
-         file, err, errcode = fetch.fetch_url(url, filename, cache)
-      end
-
+      local file, err, errcode = fetch.fetch_url(url, filename)
       fs.pop_dir()
       if not file then
          return nil, "Error fetching file: "..err, errcode
       end
-
       return file, temp_dir
    end
 end
@@ -256,22 +104,7 @@ function fetch.find_base_dir(file, temp_dir, src_url, src_dir)
    local ok, err = fs.change_dir(temp_dir)
    if not ok then return nil, err end
    fs.unpack_archive(file)
-
-   if not src_dir then
-      local rockspec = {
-         source = {
-            file = file,
-            dir = src_dir,
-            url = src_url,
-         }
-      }
-      ok, err = fetch.find_rockspec_source_dir(rockspec, ".")
-      if ok then
-         src_dir = rockspec.source.dir
-      end
-   end
-
-   local inferred_dir = src_dir or dir.deduce_base_dir(src_url)
+   local inferred_dir = src_dir or fetch.url_to_base_dir(src_url)
    local found_dir = nil
    if fs.exists(inferred_dir) then
       found_dir = inferred_dir
@@ -293,52 +126,26 @@ function fetch.find_base_dir(file, temp_dir, src_url, src_dir)
    return inferred_dir, found_dir
 end
 
-local function fetch_and_verify_signature_for(url, filename, tmpdir)
-   local sig_url = signing.signature_url(url)
-   local sig_file, err, errcode = fetch.fetch_url_at_temp_dir(sig_url, tmpdir)
-   if not sig_file then
-      return nil, "Could not fetch signature file for verification: " .. err, errcode
-   end
-
-   local ok, err = signing.verify_signature(filename, sig_file)
-   if not ok then
-      return nil, "Failed signature verification: " .. err
-   end
-
-   return fs.absolute_name(sig_file)
-end
-
 --- Obtain a rock and unpack it.
 -- If a directory is not given, a temporary directory will be created,
 -- which will be deleted on program termination.
 -- @param rock_file string: URL or filename of the rock.
 -- @param dest string or nil: if given, directory will be used as
 -- a permanent destination.
--- @param verify boolean: if true, download and verify signature for rockspec
 -- @return string or (nil, string, [string]): the directory containing the contents
 -- of the unpacked rock.
-function fetch.fetch_and_unpack_rock(url, dest, verify)
-   assert(type(url) == "string")
+function fetch.fetch_and_unpack_rock(rock_file, dest)
+   assert(type(rock_file) == "string")
    assert(type(dest) == "string" or not dest)
 
-   local name = dir.base_name(url):match("(.*)%.[^.]*%.rock")
-   local tmpname = "luarocks-rock-" .. name
-
-   local rock_file, err, errcode = fetch.fetch_url_at_temp_dir(url, tmpname, nil, true)
+   local name = dir.base_name(rock_file):match("(.*)%.[^.]*%.rock")
+   
+   local rock_file, err, errcode = fetch.fetch_url_at_temp_dir(rock_file,"luarocks-rock-"..name)
    if not rock_file then
       return nil, "Could not fetch rock file: " .. err, errcode
    end
 
-   local sig_file
-   if verify then
-      sig_file, err = fetch_and_verify_signature_for(url, rock_file, tmpname)
-      if err then
-         return nil, err
-      end
-   end
-
    rock_file = fs.absolute_name(rock_file)
-
    local unpack_dir
    if dest then
       unpack_dir = dest
@@ -347,111 +154,151 @@ function fetch.fetch_and_unpack_rock(url, dest, verify)
          return nil, "Failed unpacking rock file: " .. err
       end
    else
-      unpack_dir, err = fs.make_temp_dir(name)
-      if not unpack_dir then
-         return nil, "Failed creating temporary dir: " .. err
-      end
+      unpack_dir = fs.make_temp_dir(name)
    end
    if not dest then
       util.schedule_function(fs.delete, unpack_dir)
    end
    local ok, err = fs.change_dir(unpack_dir)
    if not ok then return nil, err end
-   ok, err = fs.unzip(rock_file)
+   ok = fs.unzip(rock_file)
    if not ok then
-      return nil, "Failed unpacking rock file: " .. rock_file .. ": " .. err
-   end
-   if sig_file then
-      ok, err = fs.copy(sig_file, ".")
-      if not ok then
-         return nil, "Failed copying signature file"
-      end
+      return nil, "Failed unpacking rock file: " .. rock_file
    end
    fs.pop_dir()
    return unpack_dir
 end
 
+function fetch.url_to_base_dir(url)
+   local base = dir.base_name(url)
+   return base:gsub("%.[^.]*$", ""):gsub("%.tar$", "")
+end
+
 --- Back-end function that actually loads the local rockspec.
 -- Performs some validation and postprocessing of the rockspec contents.
--- @param rel_filename string: The local filename of the rockspec file.
+-- @param filename string: The local filename of the rockspec file.
 -- @param quick boolean: if true, skips some steps when loading
 -- rockspec.
 -- @return table or (nil, string): A table representing the rockspec
 -- or nil followed by an error message.
-function fetch.load_local_rockspec(rel_filename, quick)
-   assert(type(rel_filename) == "string")
-   local abs_filename = fs.absolute_name(rel_filename)
+function fetch.load_local_rockspec(filename, quick)
+   assert(type(filename) == "string")
+   filename = fs.absolute_name(filename)
+   local rockspec, err = persist.load_into_table(filename)
+   if not rockspec then
+      return nil, "Could not load rockspec file "..filename.." ("..err..")"
+   end
+   if cfg.branch and (type(rockspec.source) == "table") then
+      rockspec.source.branch = cfg.branch
+   end
+   local globals = err
 
-   local basename = dir.base_name(abs_filename)
-   if basename ~= "rockspec" then
-      if not basename:match("(.*)%-[^-]*%-[0-9]*") then
+   local ok, err = true, nil
+   if not quick then
+      ok, err = type_check.type_check_rockspec(rockspec, globals)
+      if not ok then
+         return nil, filename..": "..err
+      end
+   end
+   
+   if rockspec.rockspec_format then
+      if deps.compare_versions(rockspec.rockspec_format, type_check.rockspec_format) then
+         return nil, "Rockspec format "..rockspec.rockspec_format.." is not supported, please upgrade LuaRocks."
+      end
+   end
+
+   util.platform_overrides(rockspec.build)
+   util.platform_overrides(rockspec.dependencies)
+   util.platform_overrides(rockspec.external_dependencies)
+   util.platform_overrides(rockspec.source)
+   util.platform_overrides(rockspec.hooks)
+
+   local basename = dir.base_name(filename)
+   if basename == "rockspec" then
+      rockspec.name = rockspec.package:lower()
+   else
+      rockspec.name = basename:match("(.*)-[^-]*-[0-9]*")
+      if not rockspec.name then
          return nil, "Expected filename in format 'name-version-revision.rockspec'."
       end
    end
 
-   local tbl, err = persist.load_into_table(abs_filename)
-   if not tbl then
-      return nil, "Could not load rockspec file "..abs_filename.." ("..err..")"
+   local protocol, pathname = dir.split_url(rockspec.source.url)
+   if fetch.is_basic_protocol(protocol) then
+      rockspec.source.file = rockspec.source.file or dir.base_name(rockspec.source.url)
    end
+   rockspec.source.protocol, rockspec.source.pathname = protocol, pathname
 
-   local rockspec, err = rockspecs.from_persisted_table(abs_filename, tbl, err, quick)
-   if not rockspec then
-      return nil, abs_filename .. ": " .. err
-   end
+   -- Temporary compatibility
+   if rockspec.source.cvs_module then rockspec.source.module = rockspec.source.cvs_module end
+   if rockspec.source.cvs_tag then rockspec.source.tag = rockspec.source.cvs_tag end
 
    local name_version = rockspec.package:lower() .. "-" .. rockspec.version
    if basename ~= "rockspec" and basename ~= name_version .. ".rockspec" then
       return nil, "Inconsistency between rockspec filename ("..basename..") and its contents ("..name_version..".rockspec)."
    end
 
+   rockspec.local_filename = filename
+   local filebase = rockspec.source.file or rockspec.source.url
+   local base = fetch.url_to_base_dir(filebase)
+   rockspec.source.dir = rockspec.source.dir
+                      or rockspec.source.module
+                      or ((filebase:match("%.lua$") or filebase:match("%.c$")) and ".")
+                      or base
+   if rockspec.dependencies then
+      for i = 1, #rockspec.dependencies do
+         local parsed, err = deps.parse_dep(rockspec.dependencies[i])
+         if not parsed then
+            return nil, "Parse error processing dependency '"..rockspec.dependencies[i].."': "..tostring(err)
+         end
+         rockspec.dependencies[i] = parsed
+      end
+   else
+      rockspec.dependencies = {}
+   end
+   if not quick then
+      path.configure_paths(rockspec)
+   end
+
    return rockspec
 end
 
 --- Load a local or remote rockspec into a table.
--- This is the entry point for the LuaRocks tools.
+-- This is the entry point for the LuaRocks tools. 
 -- Only the LuaRocks runtime loader should use
 -- load_local_rockspec directly.
 -- @param filename string: Local or remote filename of a rockspec.
 -- @param location string or nil: Where to download. If not given,
 -- a temporary dir is created.
--- @param verify boolean: if true, download and verify signature for rockspec
 -- @return table or (nil, string, [string]): A table representing the rockspec
 -- or nil followed by an error message and optional error code.
-function fetch.load_rockspec(url, location, verify)
-   assert(type(url) == "string")
+function fetch.load_rockspec(filename, location)
+   assert(type(filename) == "string")
 
    local name
-   local basename = dir.base_name(url)
+   local basename = dir.base_name(filename)
    if basename == "rockspec" then
       name = "rockspec"
    else
       name = basename:match("(.*)%.rockspec")
       if not name then
-         return nil, "Filename '"..url.."' does not look like a rockspec."
+         return nil, "Filename '"..filename.."' does not look like a rockspec."
       end
    end
-
-   local tmpname = "luarocks-rockspec-"..name
-   local filename, err, errcode
+   
+   local err, errcode
    if location then
       local ok, err = fs.change_dir(location)
       if not ok then return nil, err end
-      filename, err = fetch.fetch_url(url)
+      filename, err = fetch.fetch_url(filename)
       fs.pop_dir()
    else
-      filename, err, errcode = fetch.fetch_url_at_temp_dir(url, tmpname, nil, true)
+      filename, err, errcode = fetch.fetch_url_at_temp_dir(filename,"luarocks-rockspec-"..name)
    end
    if not filename then
       return nil, err, errcode
    end
 
-   if verify then
-      local _, err = fetch_and_verify_signature_for(url, filename, tmpname)
-      if err then
-         return nil, err
-      end
-   end
-
    return fetch.load_local_rockspec(filename)
 end
 
@@ -465,7 +312,7 @@ end
 -- the fetched source tarball and the temporary directory created to
 -- store it; or nil and an error message and optional error code.
 function fetch.get_sources(rockspec, extract, dest_dir)
-   assert(rockspec:type() == "rockspec")
+   assert(type(rockspec) == "table")
    assert(type(extract) == "boolean")
    assert(type(dest_dir) == "string" or not dest_dir)
 
@@ -494,62 +341,15 @@ function fetch.get_sources(rockspec, extract, dest_dir)
    if extract then
       local ok, err = fs.change_dir(store_dir)
       if not ok then return nil, err end
-      ok, err = fs.unpack_archive(rockspec.source.file)
-      if not ok then return nil, err end
-      ok, err = fetch.find_rockspec_source_dir(rockspec, ".")
-      if not ok then return nil, err end
+      fs.unpack_archive(rockspec.source.file)
+      if not fs.exists(rockspec.source.dir) then
+         return nil, "Directory "..rockspec.source.dir.." not found inside archive "..rockspec.source.file, "source.dir", source_file, store_dir
+      end
       fs.pop_dir()
    end
    return source_file, store_dir
 end
 
-function fetch.find_rockspec_source_dir(rockspec, store_dir)
-   local ok, err = fs.change_dir(store_dir)
-   if not ok then return nil, err end
-
-   local file_count, dir_count, found_dir = 0, 0, 0
-
-   if rockspec.source.dir and fs.exists(rockspec.source.dir) then
-      ok, err = true, nil
-   elseif rockspec.source.file and rockspec.source.dir then
-      ok, err = nil, "Directory "..rockspec.source.dir.." not found inside archive "..rockspec.source.file
-   elseif not rockspec.source.dir_set then -- and rockspec:format_is_at_least("3.0") then
-
-      local name = dir.base_name(rockspec.source.file or rockspec.source.url or "")
-
-      if name:match("%.lua$") or name:match("%.c$") then
-         if fs.is_file(name) then
-            rockspec.source.dir = "."
-            ok, err = true, nil
-         end
-      end
-
-      if not rockspec.source.dir then
-         for file in fs.dir() do
-            file_count = file_count + 1
-            if fs.is_dir(file) then
-               dir_count = dir_count + 1
-               found_dir = file
-            end
-         end
-
-         if dir_count == 1 then
-            rockspec.source.dir = found_dir
-            ok, err = true, nil
-         else
-            ok, err = nil, "Could not determine source directory from rock contents (" .. tostring(file_count).." file(s), "..tostring(dir_count).." dir(s))"
-         end
-      end
-   else
-      ok, err = nil, "Could not determine source directory, please set source.dir in rockspec."
-   end
-
-   fs.pop_dir()
-
-   assert(rockspec.source.dir or not ok)
-   return ok, err
-end
-
 --- Download sources for building a rock, calling the appropriate protocol method.
 -- @param rockspec table: The rockspec table
 -- @param extract boolean: When downloading compressed formats, whether to extract
@@ -560,29 +360,21 @@ end
 -- the fetched source tarball and the temporary directory created to
 -- store it; or nil and an error message.
 function fetch.fetch_sources(rockspec, extract, dest_dir)
-   assert(rockspec:type() == "rockspec")
+   assert(type(rockspec) == "table")
    assert(type(extract) == "boolean")
    assert(type(dest_dir) == "string" or not dest_dir)
 
-   -- auto-convert git://github.com URLs to use git+https
-   -- see https://github.blog/2021-09-01-improving-git-protocol-security-github/
-   if rockspec.source.url:match("^git://github%.com/")
-   or rockspec.source.url:match("^git://www%.github%.com/") then
-      rockspec.source.url = rockspec.source.url:gsub("^git://", "git+https://")
-      rockspec.source.protocol = "git+https"
-   end
-
    local protocol = rockspec.source.protocol
-   local ok, err, proto
-   if dir.is_basic_protocol(protocol) then
-      proto = fetch
+   local ok, proto
+   if fetch.is_basic_protocol(protocol) then
+      proto = require("luarocks.fetch")
    else
       ok, proto = pcall(require, "luarocks.fetch."..protocol:gsub("[+-]", "_"))
       if not ok then
          return nil, "Unknown protocol "..protocol
       end
    end
-
+   
    if cfg.only_sources_from
    and rockspec.source.pathname
    and #rockspec.source.pathname > 0 then
@@ -592,14 +384,7 @@ function fetch.fetch_sources(rockspec, extract, dest_dir)
          return nil, "Can't download "..rockspec.source.url.." -- only downloading from "..cfg.only_sources_from
       end
    end
-
-   local source_file, store_dir = proto.get_sources(rockspec, extract, dest_dir)
-   if not source_file then return nil, store_dir end
-
-   ok, err = fetch.find_rockspec_source_dir(rockspec, store_dir)
-   if not ok then return nil, err, "source.dir", source_file, store_dir end
-
-   return source_file, store_dir
+   return proto.get_sources(rockspec, extract, dest_dir)
 end
 
 return fetch
